@@ -7,28 +7,12 @@
  ******************************************************************************
  * Copyright (c) 2017-2018, Planet Labs
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  ****************************************************************************/
 
 #include "cpl_http.h"
 #include "eeda.h"
-#include "ogrgeojsonreader.h"
+#include "ogrlibjsonutils.h"
 
 #include <stdlib.h>
 #include <limits>
@@ -386,8 +370,15 @@ char **GDALEEDABaseDataset::GetBaseHTTPOptions()
                 }
             }
 
-            CPLString osServiceAccountJson(
-                CPLGetConfigOption("GOOGLE_APPLICATION_CREDENTIALS", ""));
+            CPLString osServiceAccountJson;
+            const char *pszVSIPath =
+                CSLFetchNameValue(papszOpenOptions, "VSI_PATH_FOR_AUTH");
+            if (pszVSIPath)
+                osServiceAccountJson = VSIGetPathSpecificOption(
+                    pszVSIPath, "GOOGLE_APPLICATION_CREDENTIALS", "");
+            if (osServiceAccountJson.empty())
+                osServiceAccountJson =
+                    CPLGetConfigOption("GOOGLE_APPLICATION_CREDENTIALS", "");
             if (!osServiceAccountJson.empty())
             {
                 CPLJSONDocument oDoc;
@@ -443,7 +434,8 @@ char **GDALEEDABaseDataset::GetBaseHTTPOptions()
                          "Missing EEDA_BEARER, EEDA_BEARER_FILE or "
                          "GOOGLE_APPLICATION_CREDENTIALS or "
                          "EEDA_PRIVATE_KEY/EEDA_PRIVATE_KEY_FILE + "
-                         "EEDA_CLIENT_EMAIL config option");
+                         "EEDA_CLIENT_EMAIL config option or "
+                         "VSI_PATH_FOR_AUTH open option");
                 CSLDestroy(papszOptions);
                 return nullptr;
             }
