@@ -7,23 +7,7 @@
  ******************************************************************************
  * Copyright (c) 2015, Delfim Rego <delfimrego@gmail.com>
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  ****************************************************************************/
 
 #include "safedataset.h"
@@ -636,7 +620,6 @@ SAFEDataset::~SAFEDataset()
 {
     SAFEDataset::FlushCache(true);
 
-    CPLDestroyXMLNode(psManifest);
     if (nGCPCount > 0)
     {
         GDALDeinitGCPs(nGCPCount, pasGCPList);
@@ -673,13 +656,14 @@ int SAFEDataset::CloseDependentDatasets()
 /*                      GetMetaDataObject()                             */
 /************************************************************************/
 
-CPLXMLNode *SAFEDataset::GetMetaDataObject(CPLXMLNode *psMetaDataObjects,
-                                           const char *metadataObjectId)
+const CPLXMLNode *
+SAFEDataset::GetMetaDataObject(const CPLXMLNode *psMetaDataObjects,
+                               const char *metadataObjectId)
 {
     /* -------------------------------------------------------------------- */
     /*      Look for DataObject Element by ID.                              */
     /* -------------------------------------------------------------------- */
-    for (CPLXMLNode *psMDO = psMetaDataObjects->psChild; psMDO != nullptr;
+    for (const CPLXMLNode *psMDO = psMetaDataObjects->psChild; psMDO != nullptr;
          psMDO = psMDO->psNext)
     {
         if (psMDO->eType != CXT_Element ||
@@ -706,13 +690,13 @@ CPLXMLNode *SAFEDataset::GetMetaDataObject(CPLXMLNode *psMetaDataObjects,
 /*                      GetDataObject()                                 */
 /************************************************************************/
 
-CPLXMLNode *SAFEDataset::GetDataObject(CPLXMLNode *psDataObjects,
-                                       const char *dataObjectId)
+const CPLXMLNode *SAFEDataset::GetDataObject(const CPLXMLNode *psDataObjects,
+                                             const char *dataObjectId)
 {
     /* -------------------------------------------------------------------- */
     /*      Look for DataObject Element by ID.                              */
     /* -------------------------------------------------------------------- */
-    for (CPLXMLNode *psDO = psDataObjects->psChild; psDO != nullptr;
+    for (const CPLXMLNode *psDO = psDataObjects->psChild; psDO != nullptr;
          psDO = psDO->psNext)
     {
         if (psDO->eType != CXT_Element ||
@@ -735,14 +719,15 @@ CPLXMLNode *SAFEDataset::GetDataObject(CPLXMLNode *psDataObjects,
     return nullptr;
 }
 
-CPLXMLNode *SAFEDataset::GetDataObject(CPLXMLNode *psMetaDataObjects,
-                                       CPLXMLNode *psDataObjects,
-                                       const char *metadataObjectId)
+const CPLXMLNode *
+SAFEDataset::GetDataObject(const CPLXMLNode *psMetaDataObjects,
+                           const CPLXMLNode *psDataObjects,
+                           const char *metadataObjectId)
 {
     /* -------------------------------------------------------------------- */
     /*      Look for MetadataObject Element by ID.                          */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psMDO =
+    const CPLXMLNode *psMDO =
         SAFEDataset::GetMetaDataObject(psMetaDataObjects, metadataObjectId);
 
     if (psMDO != nullptr)
@@ -792,8 +777,8 @@ int SAFEDataset::Identify(GDALOpenInfo *poOpenInfo)
     if (poOpenInfo->bIsDirectory)
     {
         VSIStatBufL sStat;
-        CPLString osMDFilename = CPLFormCIFilename(poOpenInfo->pszFilename,
-                                                   "manifest.safe", nullptr);
+        CPLString osMDFilename = CPLFormCIFilenameSafe(
+            poOpenInfo->pszFilename, "manifest.safe", nullptr);
 
         if (VSIStatL(osMDFilename, &sStat) == 0 && VSI_ISREG(sStat.st_mode))
         {
@@ -810,11 +795,14 @@ int SAFEDataset::Identify(GDALOpenInfo *poOpenInfo)
     if (poOpenInfo->nHeaderBytes < 100)
         return FALSE;
 
-    if (strstr((const char *)poOpenInfo->pabyHeader, "<xfdu:XFDU") == nullptr)
+    const char *pszHeader =
+        reinterpret_cast<const char *>(poOpenInfo->pabyHeader);
+    if (!strstr(pszHeader, "<xfdu:XFDU"))
         return FALSE;
 
-    // This driver doesn't handle Sentinel-2 data
-    if (strstr((const char *)poOpenInfo->pabyHeader, "sentinel-2") != nullptr)
+    // This driver doesn't handle Sentinel-2 or RCM (RADARSAT Constellation Mission) data
+    if (strstr(pszHeader, "sentinel-2") ||
+        strstr(pszHeader, "rcm_prod_manifest.xsd"))
         return FALSE;
 
     return TRUE;
@@ -997,40 +985,36 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
 
     if (poOpenInfo->bIsDirectory)
     {
-        osMDFilename =
-            CPLFormCIFilename(osMDFilename.c_str(), "manifest.safe", nullptr);
+        osMDFilename = CPLFormCIFilenameSafe(osMDFilename.c_str(),
+                                             "manifest.safe", nullptr);
     }
 
     /* -------------------------------------------------------------------- */
     /*      Ingest the manifest.safe file.                                  */
     /* -------------------------------------------------------------------- */
 
-    CPLXMLNode *psManifest = CPLParseXMLFile(osMDFilename);
+    auto psManifest = CPLXMLTreeCloser(CPLParseXMLFile(osMDFilename));
     if (psManifest == nullptr)
         return nullptr;
 
-    CPLString osPath(CPLGetPath(osMDFilename));
+    CPLString osPath(CPLGetPathSafe(osMDFilename));
 
     /* -------------------------------------------------------------------- */
     /*      Confirm the requested access is supported.                      */
     /* -------------------------------------------------------------------- */
     if (poOpenInfo->eAccess == GA_Update)
     {
-        CPLDestroyXMLNode(psManifest);
-        CPLError(CE_Failure, CPLE_NotSupported,
-                 "The SAFE driver does not support update access to existing"
-                 " datasets.\n");
+        ReportUpdateNotSupportedByDriver("SAFE");
         return nullptr;
     }
 
     /* -------------------------------------------------------------------- */
     /*      Get contentUnit parent element.                                 */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psContentUnits = CPLGetXMLNode(
-        psManifest, "=xfdu:XFDU.informationPackageMap.xfdu:contentUnit");
+    const CPLXMLNode *psContentUnits = CPLGetXMLNode(
+        psManifest.get(), "=xfdu:XFDU.informationPackageMap.xfdu:contentUnit");
     if (psContentUnits == nullptr)
     {
-        CPLDestroyXMLNode(psManifest);
         CPLError(CE_Failure, CPLE_OpenFailed,
                  "Failed to find <xfdu:XFDU><informationPackageMap>"
                  "<xfdu:contentUnit> in manifest file.");
@@ -1040,11 +1024,10 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Get Metadata Objects element.                                   */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psMetaDataObjects =
-        CPLGetXMLNode(psManifest, "=xfdu:XFDU.metadataSection");
+    const CPLXMLNode *psMetaDataObjects =
+        CPLGetXMLNode(psManifest.get(), "=xfdu:XFDU.metadataSection");
     if (psMetaDataObjects == nullptr)
     {
-        CPLDestroyXMLNode(psManifest);
         CPLError(CE_Failure, CPLE_OpenFailed,
                  "Failed to find <xfdu:XFDU><metadataSection>"
                  "in manifest file.");
@@ -1054,11 +1037,10 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Get Data Objects element.                                       */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psDataObjects =
-        CPLGetXMLNode(psManifest, "=xfdu:XFDU.dataObjectSection");
+    const CPLXMLNode *psDataObjects =
+        CPLGetXMLNode(psManifest.get(), "=xfdu:XFDU.dataObjectSection");
     if (psDataObjects == nullptr)
     {
-        CPLDestroyXMLNode(psManifest);
         CPLError(CE_Failure, CPLE_OpenFailed,
                  "Failed to find <xfdu:XFDU><dataObjectSection> in document.");
         return nullptr;
@@ -1069,7 +1051,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     auto poDS = std::make_unique<SAFEDataset>();
 
-    poDS->psManifest = psManifest;
+    poDS->psManifest = std::move(psManifest);
 
     /* -------------------------------------------------------------------- */
     /*      Look for "Measurement Data Unit" contentUnit elements.          */
@@ -1103,7 +1085,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
             if (*pszDataObjectID == '\0' || *pszDmdID == '\0')
                 continue;
 
-            CPLXMLNode *psDataObject =
+            const CPLXMLNode *psDataObject =
                 SAFEDataset::GetDataObject(psDataObjects, pszDataObjectID);
 
             const char *pszRepId = CPLGetXMLValue(psDataObject, "repID", "");
@@ -1127,7 +1109,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
                     continue;
 
                 // Map the metadata ID to the object element
-                CPLXMLNode *psDO = SAFEDataset::GetDataObject(
+                const CPLXMLNode *psDO = SAFEDataset::GetDataObject(
                     psMetaDataObjects, psDataObjects, pszId);
                 if (psDO == nullptr)
                     continue;
@@ -1162,9 +1144,9 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
 
             // open Annotation XML file
             const CPLString osAnnotationFilePath =
-                CPLFormFilename(osPath, pszAnnotation, nullptr);
+                CPLFormFilenameSafe(osPath, pszAnnotation, nullptr);
             const CPLString osCalibrationFilePath =
-                CPLFormFilename(osPath, pszCalibration, nullptr);
+                CPLFormFilenameSafe(osPath, pszCalibration, nullptr);
 
             CPLXMLTreeCloser psAnnotation(
                 CPLParseXMLFile(osAnnotationFilePath));
@@ -1308,15 +1290,21 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
             /* --------------------------------------------------------------------
              */
             const std::string osFullFilename(
-                CPLFormFilename(osPath, pszMeasurement, nullptr));
+                CPLFormFilenameSafe(osPath, pszMeasurement, nullptr));
 
             /* --------------------------------------------------------------------
              */
             /*      Try and open the file. */
             /* --------------------------------------------------------------------
              */
-            auto poBandFile = std::unique_ptr<GDALDataset>(
-                GDALDataset::Open(osFullFilename.c_str()));
+            std::unique_ptr<GDALDataset> poBandFile;
+            {
+                CPLTurnFailureIntoWarningBackuper oErrorsToWarnings{};
+                poBandFile = std::unique_ptr<GDALDataset>(
+                    GDALDataset::Open(osFullFilename.c_str(),
+                                      GDAL_OF_RASTER | GDAL_OF_VERBOSE_ERROR));
+            }
+
             if (poBandFile == nullptr)
             {
                 // NOP
@@ -1744,7 +1732,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Platform information                                            */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psPlatformAttrs =
+    const CPLXMLNode *psPlatformAttrs =
         SAFEDataset::GetMetaDataObject(psMetaDataObjects, "platform");
 
     if (psPlatformAttrs != nullptr)
@@ -1781,7 +1769,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Acquisition Period information                                  */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psAcquisitionAttrs =
+    const CPLXMLNode *psAcquisitionAttrs =
         SAFEDataset::GetMetaDataObject(psMetaDataObjects, "acquisitionPeriod");
 
     if (psAcquisitionAttrs != nullptr)
@@ -1802,7 +1790,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Processing information                                          */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psProcessingAttrs =
+    const CPLXMLNode *psProcessingAttrs =
         SAFEDataset::GetMetaDataObject(psMetaDataObjects, "processing");
 
     if (psProcessingAttrs != nullptr)
@@ -1816,7 +1804,7 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Measurement Orbit Reference information                         */
     /* -------------------------------------------------------------------- */
-    CPLXMLNode *psOrbitAttrs = SAFEDataset::GetMetaDataObject(
+    const CPLXMLNode *psOrbitAttrs = SAFEDataset::GetMetaDataObject(
         psMetaDataObjects, "measurementOrbitReference");
 
     if (psOrbitAttrs != nullptr)
@@ -1834,6 +1822,46 @@ GDALDataset *SAFEDataset::Open(GDALOpenInfo *poOpenInfo)
         poDS->SetMetadataItem("ORBIT_DIRECTION", pszItem);
     }
 
+    /* -------------------------------------------------------------------- */
+    /*      Footprint                                                       */
+    /* -------------------------------------------------------------------- */
+    const CPLXMLNode *psFrameSet = SAFEDataset::GetMetaDataObject(
+        psMetaDataObjects, "measurementFrameSet");
+
+    if (psFrameSet)
+    {
+        const auto psFootPrint = CPLGetXMLNode(psFrameSet, "metadataWrap."
+                                                           "xmlData."
+                                                           "safe:frameSet."
+                                                           "safe:frame."
+                                                           "safe:footPrint");
+        if (psFootPrint)
+        {
+            const char *pszSRSName =
+                CPLGetXMLValue(psFootPrint, "srsName", nullptr);
+            const char *pszCoordinates =
+                CPLGetXMLValue(psFootPrint, "gml:coordinates", nullptr);
+            if (pszSRSName &&
+                EQUAL(pszSRSName,
+                      "http://www.opengis.net/gml/srs/epsg.xml#4326") &&
+                pszCoordinates)
+            {
+                const CPLStringList aosValues(
+                    CSLTokenizeString2(pszCoordinates, " ,", 0));
+                if (aosValues.size() == 8)
+                {
+                    poDS->SetMetadataItem(
+                        "FOOTPRINT",
+                        CPLSPrintf("POLYGON((%s %s,%s %s,%s %s,%s %s, %s %s))",
+                                   aosValues[1], aosValues[0], aosValues[3],
+                                   aosValues[2], aosValues[5], aosValues[4],
+                                   aosValues[7], aosValues[6], aosValues[1],
+                                   aosValues[0]));
+                }
+            }
+        }
+    }
+
     /* -------------------------------------------------------------------- */
     /*      Initialize any PAM information.                                 */
     /* -------------------------------------------------------------------- */
